package entry

import (
	"fmt"
	"time"

	"codeberg.org/gruf/go-byteutil"
	"codeberg.org/gruf/go-kv"
	"codeberg.org/gruf/go-logger/v2/level"
	"codeberg.org/gruf/go-rtutil/v2/caller"
)

// DefaultFormatter is the global default formatter, useful
// when in case of nil Formatter, or simply wanting defaults.
var DefaultFormatter = NewTextFormatter(nil)

// Formatter provides formatting for various standard
// data types being written to an entry's format buffer.
type Formatter interface {
	// Level will append formatted level string to buffer.
	Level(buf *byteutil.Buffer, lvl level.LEVEL)

	// Timestamp will append formatted timestamp string to buffer.
	Timestamp(buf *byteutil.Buffer)

	// Caller will append formatted caller information to buffer.
	Caller(buf *byteutil.Buffer, calldepth int)

	// Fields will append formatted key-value fields to buffer.
	Fields(buf *byteutil.Buffer, fields []kv.Field)

	// Msg will append formatted msg arguments to buffer.
	Msg(buf *byteutil.Buffer, args ...interface{})

	// Msgf will append formatted msg format string with arguments to buffer.
	Msgf(buf *byteutil.Buffer, format string, args ...interface{})
}

// TextFormatter is a Levels typedef to implement the Formatter interface,
// in a manner that provides logging with timestamp, caller info and level
// prefixing log entries in plain printed text.
type TextFormatter struct {
	lvls level.Levels
}

// NewTextFormatter returns a prepared TextFormatter instance, using
// provided level.Levels for printing of level strings, else uses defaults.
func NewTextFormatter(lvls *level.Levels) Formatter {
	if lvls == nil {
		// Use default levels
		l := level.Default()
		lvls = &l
	}
	return &TextFormatter{lvls: *lvls}
}

func (f *TextFormatter) Level(buf *byteutil.Buffer, lvl level.LEVEL) {
	buf.B = append(buf.B, '[')
	buf.B = append(buf.B, f.lvls.Get(lvl)...)
	buf.B = append(buf.B, ']')
}

func (f *TextFormatter) Timestamp(buf *byteutil.Buffer) {
	const fmt = "2006-01-02 15:04:05"
	buf.B = time.Now().AppendFormat(buf.B, fmt)
}

func (f *TextFormatter) Caller(buf *byteutil.Buffer, calldepth int) {
	fn := caller.Get(calldepth + 1)
	buf.B = append(buf.B, fn...)
	buf.B = append(buf.B, `()`...)
}

func (f *TextFormatter) Fields(buf *byteutil.Buffer, fields []kv.Field) {
	kv.Fields(fields).AppendFormat(buf, false)
}

func (f *TextFormatter) Msg(buf *byteutil.Buffer, a ...interface{}) {
	buf.B = fmt.Append(buf.B, a...)
}

func (f *TextFormatter) Msgf(buf *byteutil.Buffer, s string, a ...interface{}) {
	buf.B = fmt.Appendf(buf.B, s, a...)
}

// FieldFormatter is a level.Levels typedef to implement the Formatter
// interface, in a manner that provides logging with timestamp, caller
// info and level prefixing log entries in key-value formatting.
type FieldFormatter struct {
	lvls level.Levels
}

// NewFieldFormatter returns a prepared FieldFormatter instance, using
// provided level.Levels for printing of level strings, else uses defaults.
func NewFieldFormatter(lvls *level.Levels) Formatter {
	if lvls == nil {
		// Use default levels
		l := level.Default()
		lvls = &l
	}
	return &FieldFormatter{lvls: *lvls}
}

func (f *FieldFormatter) Level(buf *byteutil.Buffer, lvl level.LEVEL) {
	buf.B = append(buf.B, `level=`...)
	buf.B = append(buf.B, f.lvls.Get(lvl)...)
}

func (f *FieldFormatter) Timestamp(buf *byteutil.Buffer) {
	const fmt = "2006-01-02 15:04:05"
	buf.B = append(buf.B, `timestamp="`...)
	buf.B = time.Now().AppendFormat(buf.B, fmt)
	buf.B = append(buf.B, '"')
}

func (f *FieldFormatter) Caller(buf *byteutil.Buffer, calldepth int) {
	fn := caller.Get(calldepth + 1)
	buf.B = append(buf.B, `func=`...)
	buf.B = append(buf.B, fn...)
	buf.B = append(buf.B, `()`...)
}

func (f *FieldFormatter) Fields(buf *byteutil.Buffer, fields []kv.Field) {
	kv.Fields(fields).AppendFormat(buf, false)
}

func (f *FieldFormatter) Msg(buf *byteutil.Buffer, a ...interface{}) {
	var vbuf byteutil.Buffer
	vbuf.B = fmt.Append(vbuf.B, a...)
	buf.B = append(buf.B, `msg=`...)
	kv.AppendQuoteValue(buf, vbuf.String())
}

func (f *FieldFormatter) Msgf(buf *byteutil.Buffer, s string, a ...interface{}) {
	var vbuf byteutil.Buffer
	vbuf.B = fmt.Append(vbuf.B, a...)
	buf.B = append(buf.B, `msg=`...)
	kv.AppendQuoteValue(buf, vbuf.String())
}
